# -*- coding: utf-8 -*-

"""This module provides access to the deployment REST api of Camunda."""

from __future__ import annotations
import datetime as dt
import dataclasses
import typing

import pycamunda.variable
import pycamunda.base
from pycamunda.request import QueryParameter, PathParameter, BodyParameter

URL_SUFFIX = '/deployment'


__all__ = ['GetList', 'Get', 'Create', 'GetResources', 'GetResource', 'Delete']


@dataclasses.dataclass
class Deployment:
    """Data class of deployment as returned by the REST api of Camunda."""
    id_: str
    name: str
    source: str
    tenant_id: str
    deployment_time: dt.datetime = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Deployment:
        deployment = cls(
            id_=data['id'],
            name=data['name'],
            source=data['source'],
            tenant_id=data['tenantId'],
        )
        if data['deploymentTime'] is not None:
            deployment.deployment_time = pycamunda.base.from_isoformat(data['deploymentTime'])

        return deployment


@dataclasses.dataclass
class Resource:
    """Data class of resource as returned by the REST api of Camunda."""
    id_: str
    name: str
    deployment_id: str

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Resource:
        return cls(
            id_=data['id'],
            name=data['name'],
            deployment_id=data['deploymentId']
        )


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    source = QueryParameter('source')
    without_source = QueryParameter('withoutSource')
    tenant_id_in = QueryParameter('tenantIdIn')
    without_tenant_id = QueryParameter('withoutTenantId', provide=pycamunda.base.value_is_true)
    include_deployments_without_tenant_id = QueryParameter(
        'includeDeploymentsWithoutTenantId', provide=pycamunda.base.value_is_true
    )
    after = QueryParameter('after')
    before = QueryParameter('before')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'id_': 'id',
            'name': 'name',
            'definition_time': 'definitionTime',
            'tenant_id': 'tenantId',
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url: str,
        id_: str = None,
        name: str = None,
        name_like: str = None,
        source: str = None,
        without_source: bool = None,
        tenant_id_in: typing.Iterable[str] = None,
        without_tenant_id: bool = False,
        include_deployments_without_tenant_id: bool = False,
        after: dt.datetime = None,
        before: dt.datetime = None,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None,
    ):
        """Get a list of deployments.
        
        :param url: Camunda Rest engine URL.
        :param id_: Filter by id.
        :param name: Filter by name.
        :param name_like: Filter by a substring of the name.
        :param source: Filter by deployment source.
        :param without_source: Filter deployments without source.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param without_tenant_id: Whether to include only deployments that belong to no tenant.
        :param include_deployments_without_tenant_id: Whether to include deployments that belong to
                                                      no tenant.
        :param after: Filter by the deployment date after the provided data.
        :param before: Filter by the deployment date before the provided data.
        :param sort_by: Sort the results by 'id', 'name', 'deployment_time' or 'tenant_id'.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.name = name
        self.name_like = name_like
        self.source = source
        self.without_source = without_source
        self.tenant_id_in = tenant_id_in
        self.without_tenant_id = without_tenant_id
        self.include_deployments_without_tenant_id = include_deployments_without_tenant_id
        self.after = after
        self.before = before
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[Deployment]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Deployment.load(deployment_json) for deployment_json in response.json())


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get a deployment.

        :param url: Camunda Rest engine url.
        :param id_: Id of the deployment.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> Deployment:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Deployment.load(response.json())


class Create(pycamunda.base.CamundaRequest):

    name = BodyParameter('deployment-name')
    enable_duplicate_filtering = BodyParameter('enable-duplicate-filtering')
    deploy_changed_only = BodyParameter('deploy-changed-only')
    source = BodyParameter('deployment-source')
    tenant_id = BodyParameter('tenant-id')

    def __init__(
        self,
        url: str,
        name: str,
        source: str = None,
        enable_duplicate_filtering: bool = False,
        deploy_changed_only: bool = False,
        tenant_id: str = None
    ):
        """Create a deployment with one or multiple resources (e.g. bpmn diagrams).

        :param url: Camunda Rest engine url
        :param name: Name of the deployment.
        :param source: Source of the deployment.
        :param enable_duplicate_filtering: Whether to check if a deployment with the same name
                                           already exists. If one does, no new deployment is done.
        :param deploy_changed_only: Whether to check if already deployed resources that are also
                                    contained in this deployment have changed. The ones that do not
                                    will not created again.
        :param tenant_id: Id of the tenant to create the deployment for.
        """
        super().__init__(url=url + URL_SUFFIX + '/create')
        self.name = name
        self.source = source
        self.enable_duplicate_filtering = enable_duplicate_filtering
        self.deploy_changed_only = deploy_changed_only
        self.tenant_id = tenant_id

        self._files = []

    def add_resource(self, file) -> None:
        """Add a resource to the deployment.

        :param file: Binary data of the resource.
        """
        self._files.append(file)

    @property
    def files(self):
        return {f'resource-{i}': resource for i, resource in enumerate(self._files)}

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        assert bool(self.files), 'Cannot create deployment without resources.'
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class GetResources(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get the resources of a deployment.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the the deployment.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/resources')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> typing.Tuple[Resource]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Resource.load(resource_json) for resource_json in response.json())


class GetResource(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    resource_id = PathParameter('resourceId')

    def __init__(self, url: str, id_: str, resource_id: str, binary: bool = False):
        """Get a resource of a deployment.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the the deployment.
        :param resource_id: Id of the resource.
        :param binary: Whether to request binary content of the resource.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/resources/{resourceId}')
        self.id_ = id_
        self.resource_id = resource_id
        self.binary = binary

    @property
    def url(self):
        return super().url + ('/data' if self.binary else '')

    def __call__(self, *args, **kwargs) -> typing.Union[Resource, typing.ByteString]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        if self.binary:
            return response.content
        return Resource.load(response.json())


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    cascade = QueryParameter('cascade')
    skip_custom_listeners = QueryParameter('skipCustomListeners')
    skip_io_mappings = QueryParameter('skipIoMappings')

    def __init__(
        self,
        url: str,
        id_: str,
        cascade: bool = False,
        skip_custom_listeners: bool = False,
        skip_io_mappings: bool = False
    ):
        """Delete a deployment.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the the deployment.
        :param cascade: Whether to cascade the action to all process instances (including historic).
        :param skip_custom_listeners: Whether to skip custom listeners and notify only builtin ones.
        :param skip_io_mappings: Whether to skip input/output mappings.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.cascade = cascade
        self.skip_custom_listeners = skip_custom_listeners
        self.skip_io_mappings = skip_io_mappings

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)



# -*- coding: utf-8 -*-

"""This module provides access to the task REST api of Camunda."""

from __future__ import annotations
import datetime as dt
import dataclasses
import enum
import typing

import pycamunda.variable
import pycamunda.base
from pycamunda.request import BodyParameter, PathParameter, QueryParameter

URL_SUFFIX = '/task'


__all__ = [
    'DelegationState', 'Get', 'GetList', 'Claim', 'Unclaim', 'Complete', 'Resolve', 'SetAssignee',
    'Delegate', 'Create', 'Update', 'LocalVariablesGet', 'LocalVariablesGetList',
    'LocalVariablesModify', 'LocalVariablesUpdate', 'LocalVariablesDelete',
    'GetCountByCandidateGroup'
]


@dataclasses.dataclass
class Task:
    """Data class of task as returned by the REST api of Camunda."""
    assignee: str
    case_definition_id: str
    case_execution_id: str
    case_instance_id: str
    delegation_state: str
    description: str
    execution_id: str
    form_key: str
    id_: str
    name: str
    owner: str
    parent_task_id: str
    priority: str
    process_definition_id: str
    process_instance_id: str
    suspended: bool
    task_definition_key: str
    created: str = None
    due: str = None
    follow_up: str = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Task:
        task = cls(
            assignee=data['assignee'],
            case_definition_id=data['caseDefinitionId'],
            case_execution_id=data['caseExecutionId'],
            case_instance_id=data['caseInstanceId'],
            delegation_state=data['delegationState'],
            description=data['description'],
            execution_id=data['executionId'],
            form_key=data['formKey'],
            id_=data['id'],
            name=data['name'],
            owner=data['owner'],
            parent_task_id=data['parentTaskId'],
            priority=data['priority'],
            process_definition_id=data['processDefinitionId'],
            process_instance_id=data['processInstanceId'],
            suspended=data['suspended'],
            task_definition_key=data['taskDefinitionKey'],
        )
        if data['created'] is not None:
            task.created = pycamunda.base.from_isoformat(data['created'])
        if data['due'] is not None:
            task.due = pycamunda.base.from_isoformat(data['due'])
        if data['followUp'] is not None:
            task.follow_up = pycamunda.base.from_isoformat(data['followUp'])

        return task


class DelegationState(enum.Enum):
    pending = 'PENDING'
    resolved = 'RESOLVED'


@dataclasses.dataclass
class IdentityLink:
    """Data class of an identity link related to user tasks as returned by the REST api of
    Camunda.
    """
    user_id: str
    group_id: str
    type_: str

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> IdentityLink:
        return cls(
            user_id=data['userId'],
            group_id=data['groupId'],
            type_=data['type']
        )


@dataclasses.dataclass
class Comment:
    """Data class of a comment that is attached to an user task."""
    id_: str
    user_id: str
    task_id: str
    message: str
    root_process_instance_id: str
    time: dt.datetime = None
    removal_time: dt.datetime = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Comment:
        comment = cls(
            id_=data['id'],
            user_id=data['userId'],
            task_id=data['taskId'],
            message=data['message'],
            root_process_instance_id=data['rootProcessInstanceId']
        )
        if data['time'] is not None:
            comment.time = pycamunda.base.from_isoformat(data['time'])
        if data['removalTime'] is not None:
            comment.removal_time = pycamunda.base.from_isoformat(data['removalTime'])

        return comment


@dataclasses.dataclass
class CountByCandidateGroup:
    """Data class of task count by candidate group."""
    group_name: typing.Optional[str]
    task_count: int

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> CountByCandidateGroup:
        return cls(
            group_name=data['groupName'],
            task_count=int(data['taskCount'])
        )


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter("id")

    def __init__(self, url: str, id_: str):
        """Get an user task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> Task:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Task.load(response.json())


class GetList(pycamunda.base.CamundaRequest):

    process_instance_id = QueryParameter('processInstanceId')
    process_instance_id_in = QueryParameter('processInstanceIdIn')
    process_instance_business_key = QueryParameter('processInstanceBusinessKey')
    process_instance_business_key_in = QueryParameter('processInstanceBusinessKeyIn')
    process_instance_business_key_like = QueryParameter('processInstanceBusinessKeyLike')
    process_definition_id = QueryParameter('processDefinitionId')
    process_definition_key = QueryParameter('processDefinitionKey')
    process_definition_key_in = QueryParameter('processDefinitionKeyIn')
    process_definition_name = QueryParameter('processDefinitionName')
    process_definition_name_like = QueryParameter('processDefinitionNameLike')
    execution_id = QueryParameter('executionId')
    case_instance_id = QueryParameter('caseInstanceId')
    case_instance_business_key = QueryParameter('caseInstanceBusinessKey')
    case_instance_business_key_like = QueryParameter('caseInstanceBusinessKeyLike')
    case_definition_id = QueryParameter('caseDefinitionId')
    case_definition_key = QueryParameter('caseDefinitionKey')
    case_instance_name = QueryParameter('caseInstanceName')
    case_instance_name_like = QueryParameter('caseInstanceNameLike')
    case_execution_id = QueryParameter('caseExecutionId')
    activity_instance_id_in = QueryParameter('activityInstanceIdIn')
    tenant_id_in = QueryParameter('tenantIdIn')
    without_tenant_id = QueryParameter('withoutTenantId', provide=pycamunda.base.value_is_true)
    assignee = QueryParameter('assignee')
    assignee_like = QueryParameter('assigneeLike')
    assignee_in = QueryParameter('assigneeIn')
    owner = QueryParameter('owner')
    candidate_group = QueryParameter('candidateGroup')
    candidate_user = QueryParameter('candidateUser')
    include_assigned_tasks = QueryParameter('includeAssignedTasks')
    involved_user = QueryParameter('involvedUser')
    assigned = QueryParameter('assigned')
    unassigned = QueryParameter('unassigned')
    task_definition_key = QueryParameter('taskDefinitionKey')
    task_definition_key_in = QueryParameter('taskDefinitionKeyIn')
    task_definition_key_like = QueryParameter('taskDefinitionKeyLike')
    name = QueryParameter('name')
    name_not_equal = QueryParameter('nameNotEqual')
    name_like = QueryParameter('nameLike')
    name_not_like = QueryParameter('nameNotLike')
    description = QueryParameter('description')
    description_like = QueryParameter('descriptionLike')
    priority = QueryParameter('priority')
    max_priority = QueryParameter('maxPriority')
    min_priority = QueryParameter('minPriority')
    due_date = QueryParameter('dueDate')
    due_after = QueryParameter('dueAfter')
    due_before = QueryParameter('dueBefore')
    follow_up_date = QueryParameter('followUpDate')
    follow_up_after = QueryParameter('followUpAfter')
    follow_up_before = QueryParameter('followUpBefore')
    follow_up_before_or_not_existent = QueryParameter('followUpBeforeOrNotExistent')
    created_on = QueryParameter('createdOn')
    created_after = QueryParameter('createdAfter')
    created_before = QueryParameter('createdBefore')
    delegation_state = QueryParameter('delegationState')
    candidate_groups = QueryParameter('candidateGroups')
    with_candidate_groups = QueryParameter(
        'withCandidateGroups', provide=pycamunda.base.value_is_true
    )
    without_candidate_groups = QueryParameter(
        'withoutCandidateGroups', provide=pycamunda.base.value_is_true
    )
    with_candidate_users = QueryParameter(
        'withCandidateUsers', provide=pycamunda.base.value_is_true
    )
    without_candidate_users = QueryParameter(
        'withoutCandidateUsers', provide=pycamunda.base.value_is_true
    )
    active = QueryParameter('active', provide=pycamunda.base.value_is_true)
    suspended = QueryParameter('suspended', provide=pycamunda.base.value_is_true)
    # TODO add variables parameters
    parent_task_id = QueryParameter('parentTaskId')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'instance_id': 'instanceId',
            'case_instance_id': 'caseInstanceId',
            'due_date': 'dueDate',
            'execution_id': 'executionId',
            'case_execution_id': 'caseExecutionId',
            'assignee': 'assignee',
            'created': 'created',
            'description': 'description',
            'id': 'id',
            'name': 'name',
            'name_case_insensitive': 'nameCaseInsensitive',
            'priority': 'priority'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')
    # TODO add expression parameters, consider explicit addition of each possible expression function

    def __init__(
        self,
        url: str,
        process_instance_id: str = None,
        process_instance_id_in: typing.Iterable[str] = None,
        process_instance_business_key: str = None,
        process_instance_business_key_in: typing.Iterable[str] = None,
        process_instance_business_key_like: str = None,
        process_definition_id: str = None,
        process_definition_key: str = None,
        process_definition_key_in: typing.Iterable[str] = None,
        process_definition_name: str = None,
        process_definition_name_like: str = None,
        execution_id: str = None,
        case_instance_id: str = None,
        case_instance_business_key: str = None,
        case_instance_business_key_like: str = None,
        case_definition_id: str = None,
        case_definition_key: str = None,
        case_instance_name: str = None,
        case_instance_name_like: str = None,
        case_execution_id: str = None,
        activity_instance_id_in: typing.Iterable[str] = None,
        tenant_id_in: typing.Iterable[str] = None,
        without_tenant_id: bool = False,
        assignee: str = None,
        assignee_like: str = None,
        assignee_in: typing.Iterable[str] = None,
        owner: str = None,
        candidate_group: str = None,
        candidate_user: str = None,
        include_assigned_tasks: bool = None,
        involved_user: str = None,
        assigned: bool = None,
        unassigned: bool = None,
        task_definition_key: str = None,
        task_definition_key_in: typing.Iterable[str] = None,
        task_definition_key_like: str = None,
        name: str = None,
        name_not_equal: str = None,
        name_like: str = None,
        name_not_like: str = None,
        description: str = None,
        description_like: str = None,
        priority: int = None,
        max_priority: int = None,
        min_priority: int = None,
        due_date: dt.datetime = None,
        due_after: dt.datetime = None,
        due_before: dt.datetime = None,
        follow_up_date: dt.datetime = None,
        follow_up_after: dt.datetime = None,
        follow_up_before: dt.datetime = None,
        follow_up_before_or_not_existent: dt.datetime = None,
        created_on: dt.datetime = None,
        created_after: dt.datetime = None,
        created_before: dt.datetime = None,
        delegation_state: typing.Union[str, DelegationState] = None,
        candidate_groups: typing.Iterable['str'] = None,
        with_candidate_groups: bool = False,
        without_candidate_groups: bool = False,
        with_candidate_users: bool = False,
        without_candidate_users: bool = False,
        active: bool = False,
        suspended: bool = False,
        parent_task_id: str = None,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None
    ):
        """Get a list of user tasks.

        :param url: Camunda Rest engine URL.
        :param process_instance_id: Filter by the process instance id.
        :param process_instance_id_in: Filter whether the process instance id is one of mutliple
                                       ones.
        :param process_instance_business_key: Filter by the business key of the process instance.
        :param process_instance_business_key_in: Filter whether the process instance business key is
                                                 one of multiple ones.
        :param process_instance_business_key_like: Filter by a substring of the process instance
                                                   business key.
        :param process_definition_id: Filter by the process definition id.
        :param process_definition_key: Filter by the process definition key.
        :param process_definition_key_in: Filter whether the the process definition key is one of
                                          multiple ones.
        :param process_definition_name: Filter by the process definition name.
        :param process_definition_name_like: Filter by a substring of the process definition name.
        :param execution_id: Filter by execution id.
        :param case_instance_id: Filter by case instance id.
        :param case_instance_business_key: Filter by case instance business key.
        :param case_instance_business_key_like: Filter by a substring of the case instance business
                                                key.
        :param case_definition_id: Filter by case definition id.
        :param case_definition_key: Filter by case definition key.
        :param case_instance_name: Filter by case instance name.
        :param case_instance_name_like: Filter by a substring of the case instance name.
        :param case_execution_id: Filter by case execution id.
        :param activity_instance_id_in: Filter whether the activity instance id is one of multiple
                                        ones.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param without_tenant_id: Whether to include only tasks that belong to no tenant.
        :param assignee: Filter by assignee.
        :param assignee_like: Filter by a substring of the assignee.
        :param assignee_in: Filter whether the assignee is one of multiple ones.
        :param owner: Filter by owner.
        :param candidate_group: Filter by candidate group.
        :param candidate_user: Filter by candidate user.
        :param include_assigned_tasks: Include only tasks that are assigned to an user.
        :param involved_user: Filter whether a user is involved in the task in some way.
                              (e.g. assigned)
        :param assigned: Filter whether the task is assigned.
        :param unassigned: Filter whether the task is unassigned.
        :param task_definition_key: Filter by task definition key.
        :param task_definition_key_in: Filter whether the task definition key is one of multiple
                                       ones.
        :param task_definition_key_like: Filter by a substring of the task definition key.
        :param name: Filter by name.
        :param name_not_equal: Filter tasks whose names are not equal the provided value.
        :param name_like: Filter by a substring of the name.
        :param name_not_like: Filter tasks whose names do not have the provided value as substring.
        :param description: Filter by description.
        :param description_like: Filter by a substring of the description.
        :param priority: Filter by priority.
        :param max_priority: Filter tasks whose priority is less than or equal the provided value.
        :param min_priority: Filter tasks whose priority is greater than or equal the provided
                             value.
        :param due_date: Filter by due date.
        :param due_after: Include only tasks whose due date expires after the provided date.
        :param due_before: Include only tasks whose due date expires before the provided date.
        :param follow_up_date: Filter by follow up date.
        :param follow_up_after: Include only tasks whose follow up date expires after the provided
                                date.
        :param follow_up_before: Include only tasks whose follow up date expires before the provided
                                 date.
        :param follow_up_before_or_not_existent: Include only tasks whose follow up date expires
                                                 before the provided date or have no follow up date.
        :param created_on: Filter by the creation date.
        :param created_after: Include only tasks whose creation date expires after the provided
                              date.
        :param created_before: Include only tasks whose creation date expires before the provided
                               date.
        :param delegation_state: Filter by delegation state. Valid values are 'PENDING' and
                                 'RESOLVED'.
        :param candidate_groups: Filter whether the candidate group is one of multiple ones.
        :param with_candidate_groups: Filter whether the task has a candidate group.
        :param without_candidate_groups: Filter whether the task has no candidate group.
        :param with_candidate_users: Filter whether the task has candidate users.
        :param without_candidate_users: Filter whether the nas no candidate users.
        :param active: Filter whether the task is active.
        :param suspended: Filter whether the is suspended.
        :param parent_task_id: Filter by parent task id.
        :param sort_by: Sort the results by 'instance_id', 'case_instance_id', 'due_date',
                        'execution_id', 'case_execution_id', 'assignee', 'created', 'description',
                        'id', 'name', 'name_case_insensitive' or 'priority'.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.process_instance_id = process_instance_id
        self.process_instance_id_in = process_instance_id_in
        self.process_instance_business_key = process_instance_business_key
        self.process_instance_business_key_in = process_instance_business_key_in
        self.process_instance_business_key_like = process_instance_business_key_like
        self.process_definition_id = process_definition_id
        self.process_definition_key = process_definition_key
        self.process_definition_key_in = process_definition_key_in
        self.process_definition_name = process_definition_name
        self.process_definition_name_like = process_definition_name_like
        self.execution_id = execution_id
        self.case_instance_id = case_instance_id
        self.case_instance_business_key = case_instance_business_key
        self.case_instance_business_key_like = case_instance_business_key_like
        self.case_definition_id = case_definition_id
        self.case_definition_key = case_definition_key
        self.case_instance_name = case_instance_name
        self.case_instance_name_like = case_instance_name_like
        self.case_execution_id = case_execution_id
        self.activity_instance_id_in = activity_instance_id_in
        self.tenant_id_in = tenant_id_in
        self.without_tenant_id = without_tenant_id
        self.assignee = assignee
        self.assignee_like = assignee_like
        self.assignee_in = assignee_in
        self.owner = owner
        self.candidate_group = candidate_group
        self.candidate_user = candidate_user
        self.include_assigned_tasks = include_assigned_tasks
        self.involved_user = involved_user
        self.assigned = assigned
        self.unassigned = unassigned
        self.task_definition_key = task_definition_key
        self.task_definition_key_in = task_definition_key_in
        self.task_definition_key_like = task_definition_key_like
        self.name = name
        self.name_not_equal = name_not_equal
        self.name_like = name_like
        self.name_not_like = name_not_like
        self.description = description
        self.description_like = description_like
        self.priority = priority
        self.min_priority = min_priority
        self.max_priority = max_priority
        self.due_date = due_date
        self.due_after = due_after
        self.due_before = due_before
        self.follow_up_date = follow_up_date
        self.follow_up_after = follow_up_after
        self.follow_up_before = follow_up_before
        self.follow_up_before_or_not_existent = follow_up_before_or_not_existent
        self.created_on = created_on
        self.created_after = created_after
        self.created_before = created_before
        self.delegation_state = None
        if delegation_state is not None:
            self.delegation_state = DelegationState(delegation_state)
        self.candidate_groups = candidate_groups
        self.with_candidate_groups = with_candidate_groups
        self.without_candidate_groups = without_candidate_groups
        self.with_candidate_users = with_candidate_users
        self.without_candidate_users = without_candidate_users
        self.active = active
        self.suspended = suspended
        self.parent_task_id = parent_task_id
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[Task]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Task.load(task_json) for task_json in response.json())


class Claim(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    user_id = BodyParameter('userId')

    def __init__(self, url: str, id_: str, user_id: str):
        """Claim a user task for a specific user. Only tasks that are not already claimed by other
        users can be claimed. To change the assignee of a task independently on whether it is
        already claimed by an user, the class 'SetAssignee' can be used.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user task.
        :param user_id: Id of the user to set as assignee for the task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/claim')
        self.id_ = id_
        self.user_id = user_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class Unclaim(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Unclaim an user task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/unclaim')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class Complete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    variables = BodyParameter('variables')
    with_variables_in_return = BodyParameter('withVariablesInReturn')

    def __init__(self, url: str, id_: str, with_variables_in_return: bool = False):
        """Complete an user task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/complete')
        self.id_ = id_
        self.with_variables_in_return = with_variables_in_return

        self.variables = {}

    def add_variable(
        self, name: str, value: typing.Any, type_: str = None, value_info: typing.Any = None
    ) -> None:
        """Add a variable to send to the Camunda process instance.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.variables[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

    def __call__(
        self, *args, **kwargs
    ) -> typing.Optional[typing.Dict[str, pycamunda.variable.Variable]]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        if self.with_variables_in_return:
            return {
                name: pycamunda.variable.Variable.load(var_json)
                for name, var_json in response.json().items()
            }


class Resolve(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    variables = BodyParameter('variables')

    def __init__(self, url: str, id_: str):
        """Resolve an user task that was delegated to the current assignee and send it back to the
        original owner. It is necessary that the task was delegated. The assignee of the user task
        will be set back to the owner of the task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/resolve')
        self.id_ = id_

        self.variables = {}

    def add_variable(
        self, name: str, value: typing.Any, type_: str = None, value_info: typing.Any = None
    ) -> None:
        """Add a variable to send to the Camunda process instance.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.variables[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class SetAssignee(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    user_id = BodyParameter('userId')

    def __init__(self, url: str, id_: str, user_id: str):
        """Set the assignee for an user task. Overwrites existing assignees.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user task.
        :param user_id: Id of the user to set as assignee for the task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/assignee')
        self.id_ = id_
        self.user_id = user_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class Delegate(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    user_id = BodyParameter('userId')

    def __init__(self, url: str, id_: str, user_id: str):
        """Delegate an user task to an user.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user task.
        :param user_id: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/delegate')
        self.id_ = id_
        self.user_id = user_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class Create(pycamunda.base.CamundaRequest):

    id_ = BodyParameter('id')
    name = BodyParameter('name')
    description = BodyParameter('description')
    assignee = BodyParameter('assignee')
    owner = BodyParameter('owner')
    delegation_state = BodyParameter('delegationState')
    due = BodyParameter('due')
    follow_up = BodyParameter('followUp')
    priority = BodyParameter('priority')
    parent_task_id = BodyParameter('parentTaskId')
    case_instance_id = BodyParameter('caseInstanceId')
    tenant_id = BodyParameter('tenantId')

    def __init__(
        self,
        url: str,
        id_: str,
        name: str = None,
        description: str = None,
        assignee: str = None,
        owner: str = None,
        delegation_state: typing.Union[str, DelegationState] = None,
        due: dt.datetime = None,
        follow_up: dt.datetime = None,
        priority: int = None,
        parent_task_id: str = None,
        case_instance_id: str = None,
        tenant_id: str = None
    ):
        """Create an user task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the task.
        :param name: Name of the task.
        :param description: Description of the task.
        :param assignee: The user the task is assigned to.
        :param owner: The owner of the task.
        :param delegation_state: The delegation state. Valid values are 'RESOLVED' and 'PENDING'.
        :param due: Due date of the task.
        :param follow_up: Follow up date of the task.
        :param priority: Priority of the task.
        :param parent_task_id: Id of the parent task in case this task is a subtask.
        :param case_instance_id: Id of the case instance.
        :param tenant_id: Id of the tenant.
        """
        super().__init__(url=url + URL_SUFFIX + '/create')
        self.id_ = id_
        self.name = name
        self.description = description
        self.assignee = assignee
        self.owner = owner
        self.delegation_state = None
        if delegation_state is not None:
            self.delegation_state = DelegationState(delegation_state)
        self.due = due
        self.follow_up = follow_up
        self.priority = priority
        self.parent_task_id = parent_task_id
        self.case_instance_id = case_instance_id
        self.tenant_id = tenant_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class Update(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    name = BodyParameter('name')
    description = BodyParameter('description')
    assignee = BodyParameter('assignee')
    owner = BodyParameter('owner')
    delegation_state = BodyParameter('delegationState')
    due = BodyParameter('due')
    follow_up = BodyParameter('followUp')
    priority = BodyParameter('priority')
    parent_task_id = BodyParameter('parentTaskId')
    case_instance_id = BodyParameter('caseInstanceId')
    tenant_id = BodyParameter('tenantId')

    def __init__(
        self,
        url: str,
        id_: str,
        name: str,
        description: str,
        assignee: str,
        owner: str,
        delegation_state: typing.Union[str, DelegationState],
        due: dt.datetime,
        follow_up: dt.datetime,
        priority: int,
        parent_task_id: str,
        case_instance_id: str,
        tenant_id: str
    ):
        """Update an user task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the task.
        :param name: Name of the task.
        :param description: Description of the task.
        :param assignee: The user the task is assigned to.
        :param owner: The owner of the task.
        :param delegation_state: The delegation state. Valid values are 'RESOLVED' and 'PENDING'.
        :param due: Due date of the task.
        :param follow_up: Follow up date of the task.
        :param priority: Priority of the task.
        :param parent_task_id: Id of the parent task in case this task is a subtask.
        :param case_instance_id: Id of the case instance.
        :param tenant_id: Id of the tenant. Cannot be changed. Has to be the same value as the task
                          already has.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.name = name
        self.description = description
        self.assignee = assignee
        self.owner = owner
        self.delegation_state = None
        if delegation_state is not None:
            self.delegation_state = DelegationState(delegation_state)
        self.due = due
        self.follow_up = follow_up
        self.priority = priority
        self.parent_task_id = parent_task_id
        self.case_instance_id = case_instance_id
        self.tenant_id = tenant_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class IdentityLinksGetList(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    type_ = QueryParameter('type')

    def __init__(self, url: str, task_id: str = None, type_: str = None) -> None:
        """Get the identity links of an user task.

        An identity link is a relationship between an user task and an user or a group. E.g. when
        the user is the assignee / owner or one of the candidate users of the task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param type_: Type of the identity link. Can be any custom string. Pre-defined types are
                      'assignee' (user), 'owner' (user) and 'candidate' (user / group).
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/identity-links')
        self.task_id = task_id
        self.type_ = type_

    def __call__(self, *args, **kwargs) -> typing.Tuple[IdentityLink]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(IdentityLink.load(link_json) for link_json in response.json())


class IdentityLinksAdd(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    user_id = BodyParameter('userId')
    group_id = BodyParameter('groupId')
    type_ = BodyParameter('type')

    def __init__(
        self, url: str, task_id: str, type_: str, user_id: str = None, group_id: str = None
    ):
        """Add an identity link to an user task.

        An identity link is a relationship between an user task and an user or a group. E.g. when
        the user is the assignee / owner or one of the candidate users of the task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param user_id: Id of the user. Can not be provided if group_id is provided.
        :param group_id: Id of the groupt. Can not be provided if user_id is provided.
        :param type_: Type of the identity link. Can be any custom string. Pre-defined types are
                      'assignee' (user), 'owner' (user) and 'candidate' (user / group).
        """
        assert (user_id is None) != (group_id is None), (
            'Either \'user_id\' or \'group_id\' has to be provided, not both.'
        )
        super().__init__(url=url + URL_SUFFIX + '/{id}/identity-links')
        self.task_id = task_id
        self.user_id = user_id
        self.group_id = group_id
        self.type_ = type_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class IdentityLinksDelete(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    user_id = BodyParameter('userId')
    group_id = BodyParameter('groupId')
    type_ = BodyParameter('type')

    def __init__(
        self, url: str, task_id: str, type_: str, user_id: str = None, group_id: str = None
    ):
        """Delete an identity link of an user task.

        An identity link is a relationship between an user task and an user or a group. E.g. when
        the user is the assignee / owner or one of the candidate users of the task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param user_id: Id of the user. Can not be provided if group_id is provided.
        :param group_id: Id of the groupt. Can not be provided if user_id is provided.
        :param type_: Type of the identity link. Can be any custom string. Pre-defined types are
                      'assignee' (user), 'owner' (user) and 'candidate' (user / group).
        """
        assert (user_id is None) != (group_id is None), (
            'Either \'user_id\' or \'group_id\' has to be provided, not both.'
        )
        super().__init__(url=url + URL_SUFFIX + '/{id}/identity-links/delete')
        self.task_id = task_id
        self.user_id = user_id
        self.group_id = group_id
        self.type_ = type_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class CommentGetList(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')

    def __init__(self, url: str, task_id: str) -> None:
        """Get the comments for an user task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/comment')
        self.task_id = task_id

    def __call__(self, *args, **kwargs) -> typing.Tuple[Comment]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Comment.load(data=comment_json) for comment_json in response.json())


class CommentGet(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    comment_id = PathParameter('commentId')

    def __init__(self, url: str, task_id: str, comment_id: str) -> None:
        """Get a comment for an user task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param comment_id: Id of the comment.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/comment/{commentId}')
        self.task_id = task_id
        self.comment_id = comment_id

    def __call__(self, *args, **kwargs) -> Comment:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Comment.load(data=response.json())


class CommentCreate(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    message = BodyParameter('message')

    def __init__(self, url: str, task_id: str, message: str) -> None:
        """Create a comment for an user task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param message: Message of the task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/comment/create')
        self.task_id = task_id
        self.message = message

    def __call__(self, *args, **kwargs) -> Comment:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return Comment.load(data=response.json())


class LocalVariablesGet(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    var_name = PathParameter('varName')
    deserialize_value = QueryParameter('deserializeValue')

    def __init__(
        self,
        url: str,
        task_id: str,
        var_name: str,
        deserialize_value: bool = False,
        binary: bool = False
    ):
        """Get a local variable of an user task.

        Local variables are variables that do only exist in the context of a task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param var_name: Name of the variable.
        :param deserialize_value: Whether serializable variable values are deserialized on server
                                  side.
        :param binary: Whether the requested variable is a binary array or file variable.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/localVariables/{varName}')
        self.task_id = task_id
        self.var_name = var_name
        self.deserialize_value = deserialize_value
        self.binary = binary

    @property
    def url(self):
        return super().url + ('/data' if self.binary else '')

    def __call__(self, *args, **kwargs) -> pycamunda.variable.Variable:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        if self.binary:
            return response.content
        return pycamunda.variable.Variable.load(data=response.json())


class LocalVariablesGetList(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    deserialize_values = QueryParameter('deserializeValues')

    def __init__(
        self,
        url: str,
        task_id: str,
        deserialize_values: bool = False
    ):
        """Get local variables of an user task.

        Local variables are variables that do only exist in the context of a task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param deserialize_values: Whether serializable variable values are deserialized on server
                                   side.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/localVariables')
        self.task_id = task_id
        self.deserialize_values = deserialize_values

    def __call__(self, *args, **kwargs) -> typing.Dict[str, pycamunda.variable.Variable]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return {
            name: pycamunda.variable.Variable.load(data=var_json)
            for name, var_json in response.json().items()
        }


class LocalVariablesModify(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    modifications = BodyParameter('modifications')
    deletions = BodyParameter('deletions')

    def __init__(self, url: str, task_id: str, deletions: typing.Iterable[str] = None):
        """Modify local variables of an user task. This can be either updating or deleting
        variables.

        Local variables are variables that do only exist in the context of a task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param deletions: Variables to delete.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/localVariables')
        self.task_id = task_id
        self.deletions = deletions

        self.modifications = {}

    def add_variable(
            self, name: str, value: typing.Any, type_: str = None, value_info: typing.Any = None
    ) -> None:
        """Add a variable to modify.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.modifications[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

    def body_parameters(self, apply: typing.Callable = ...) -> typing.Dict[str, typing.Any]:
        params = super().body_parameters(apply=apply)
        deletions = params.get('deletions', [])
        if isinstance(deletions, str):
            params['deletions'] = [deletions]
        else:
            params['deletions'] = list(deletions)
        return params

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class LocalVariablesUpdate(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    var_name = PathParameter('varName')
    value = BodyParameter('value')
    type_ = BodyParameter('type')
    value_info = BodyParameter('valueInfo')

    def __init__(
        self,
        url: str,
        task_id: str,
        var_name: str,
        value: typing.Any, type_: str = None,
        value_info: typing.Any = None
    ):
        """Update a local variable. May be used with binary and file variables.

        Local variables are variables that do only exist in the context of a task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param var_name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable. To send binary variables use the value 'Bytes' and
                      to send the binary value of a file variable use the value 'File' for this
                      parameter.
        :param value_info: Additional information regarding the value type. Ignored if 'type_' is
                           'Bytes' or 'File'.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/localVariables/{varName}')
        self.task_id = task_id
        self.var_name = var_name
        self.value = value
        self.type_ = type_
        self.value_info = value_info

    def _is_binary(self):
        return self.type_ in ('File', 'Bytes')

    @property
    def url(self):
        return super().url + ('/data' if self._is_binary() else '')

    def body_parameters(self, apply: typing.Callable = ...):
        if self._is_binary():
            return {'valueType': self.type_}
        return super().body_parameters(apply=apply)

    @property
    def files(self):
        return {'data': self.value}

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        if self._is_binary():
            response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)
        else:
            response = super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)

        if not response:
            pycamunda.base._raise_for_status(response)


class LocalVariablesDelete(pycamunda.base.CamundaRequest):

    task_id = PathParameter('id')
    var_name = PathParameter('varName')

    def __init__(self, url: str, task_id: str, var_name: str):
        """Delete a local variable.

        Local variables are variables that do only exist in the context of a task.

        :param url: Camunda Rest engine URL.
        :param task_id: Id of the task.
        :param var_name: Name of the variable.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/localVariables/{varName}')
        self.task_id = task_id
        self.var_name = var_name

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class GetCountByCandidateGroup(pycamunda.base.CamundaRequest):

    def __init__(self, url: str):
        """Get the number of tasks for each candidate group.

        :param url: Camunda Rest engine URL.
        """
        super().__init__(url=url + URL_SUFFIX + '/report/candidate-group-count')

    def __call__(self, *args, **kwargs) -> typing.Tuple[CountByCandidateGroup]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(CountByCandidateGroup.load(data=count_json) for count_json in response.json())



# -*- coding: utf-8 -*-

"""This module provides access to the variable instance REST api of Camunda"""

from __future__ import annotations
import dataclasses
import typing

import pycamunda.base
from pycamunda.request import QueryParameter, PathParameter

URL_SUFFIX = '/variable-instance'


__all__ = ['GetList', 'Get']


@dataclasses.dataclass
class Variable:
    """Data class of variable as returned by the REST api of Camunda."""
    value: typing.Any
    type_: str
    value_info: typing.Dict
    local: bool = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Variable:
        variable = cls(
            value=data['value'],
            type_=data['type'],
            value_info=data['valueInfo']
        )
        try:
            variable.local = data['local']
        except KeyError:
            pass
        return variable


@dataclasses.dataclass
class VariableInstance:
    """Data class of variable instance as returned by the REST api of Camunda."""
    id_: str
    name: str
    type_: str
    value: typing.Any
    value_info: typing.Dict
    process_instance_id: str
    execution_id: str
    case_instance_id: str
    case_execution_id: str
    task_id: str
    activity_instance_id: str
    tenant_id: str
    error_message: str

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> VariableInstance:
        return cls(
            id_=data['id'],
            name=data['name'],
            type_=data['type'],
            value=data['value'],
            value_info=data['valueInfo'],
            process_instance_id=data['processInstanceId'],
            execution_id=data['executionId'],
            case_instance_id=data['caseInstanceId'],
            case_execution_id=data['caseExecutionId'],
            task_id=data['taskId'],
            activity_instance_id=data['activityInstanceId'],
            tenant_id=data['tenantId'],
            error_message=data['errorMessage']
        )


class GetList(pycamunda.base.CamundaRequest):

    name = QueryParameter('variableName')
    name_like = QueryParameter('variableNameLike')
    process_instance_id_in = QueryParameter('processInstanceId')
    execution_id_in = QueryParameter('executionIdIn')
    case_instance_id_in = QueryParameter('caseInstanceIdIn')
    case_execution_id_in = QueryParameter('caseExecutionIdIn')
    task_id_in = QueryParameter('taskIdIn')
    activity_instance_id_in = QueryParameter('activityInstanceIdIn')
    tenant_id_in = QueryParameter('tenantIdIn')
    variable_values = QueryParameter('variableValues')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'name': 'variableName',
            'type_': 'variableType',
            'activity_instance_id': 'activityInstanceIdIn',
            'tenant_id': 'tenantId'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')
    deserialize_values = QueryParameter('deserializeValues')

    def __init__(
        self,
        url: str,
        name: str = None,
        name_like: str = None,
        process_instance_id_in: typing.Iterable[str] = None,
        case_instance_id_in: typing.Iterable[str] = None,
        case_execution_id_in: typing.Iterable[str] = None,
        task_id_in: typing.Iterable[str] = None,
        activity_instance_id_in: typing.Iterable[str] = None,
        tenant_id_in: typing.Iterable[str] = None,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None,
        deserialize_values: bool = False
    ):
        """Get a list of variable instances.

        :param url: Camunda Rest engine URL.
        :param name: Filter by variable name.
        :param name_like: Filter by a substring of the variable name.
        :param process_instance_id_in: Filter whether the process instance id is one of multiple
                                       ones.
        :param case_instance_id_in: Filter whether the case instance id is one of multiple ones.
        :param case_execution_id_in: Filter whether the case execution id is one of multiple ones.
        :param task_id_in: Filter whether the task id is one of multiple ones.
        :param activity_instance_id_in: Filter whether the activity instance id is one of multiple
                                        ones.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param sort_by: Sort the results by 'name', 'type_', 'activity_instance_id' or 'tenant_id'.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        :param deserialize_values: Whether serializable variable values are deserialized on server
                                   side.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.name = name
        self.name_like = name_like
        self.process_instance_id_in = process_instance_id_in
        self.case_instance_id_in = case_instance_id_in
        self.case_execution_id_in = case_execution_id_in
        self.task_id_in = task_id_in
        self.activity_instance_id_in = activity_instance_id_in
        self.tenant_id_in = tenant_id_in
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results
        self.deserialize_values = deserialize_values

        self.variable_values = []

    def _add_value_filter(self, name: str, criteria: str, value: str) -> None:
        """Add a filter to include only variables with certain values.

        :param name: Name of the variable.
        :param criteria: Filter criteria. Valid values are 'eq' for equal, 'neq' for not equal,
                         'gt' for greater than, 'gteq' for greater than or equal, 'lt' for less
                         than, 'lteq' for less than or equal and 'like'.
        :param value: Value to filter for.
        """
        self.variable_values.append((name, criteria, str(value)))

    def add_equal_value_filter(self, name: str, value: typing.Any) -> None:
        """Add a filter to include only variables with values equal a provided value.

        :param name: Name of the variable.
        :param value: Value to filter for.
        """
        self._add_value_filter(name=name, criteria='eq', value=value)

    def add_not_equal_value_filter(self, name: str, value: typing.Any) -> None:
        """Add a filter to include only variables with values not equal a provided value.

        :param name: Name of the variable.
        :param value: Value to filter for.
        """
        self._add_value_filter(name=name, criteria='neq', value=value)

    def add_greater_than_value_filter(self, name: str, value: typing.Any) -> None:
        """Add a filter to include only variables with values greater than a provided value.

        :param name: Name of the variable.
        :param value: Value to filter for.
        """
        self._add_value_filter(name=name, criteria='gt', value=value)

    def add_greater_than_equal_value_filter(self, name: str, value: typing.Any) -> None:
        """Add a filter to include only variables with values greater than or equal a provided
        value.

        :param name: Name of the variable.
        :param value: Value to filter for.
        """
        self._add_value_filter(name=name, criteria='gteq', value=value)

    def add_less_than_value_filter(self, name: str, value: typing.Any) -> None:
        """Add a filter to include only variables with values less than a provided value.

        :param name: Name of the variable.
        :param value: Value to filter for.
        """
        self._add_value_filter(name=name, criteria='lt', value=value)

    def add_less_than_equal_value_filter(self, name: str, value: typing.Any) -> None:
        """Add a filter to include only variables with values less than or equal a provided value.

        :param name: Name of the variable.
        :param value: Value to filter for.
        """
        self._add_value_filter(name=name, criteria='lteq', value=value)

    def add_like_value_filter(self, name: str, value: typing.Any) -> None:
        """Add a filter to include only variables with values like a provided value.

        :param name: Name of the variable.
        :param value: Value to filter for.
        """
        self._add_value_filter(name=name, criteria='like', value=value)

    def query_parameters(self, apply: typing.Callable = ...):
        params = super().query_parameters(apply=apply)
        if params['variableValues']:
            params['variableValues'] = ','.join(
                ('_'.join(var_val)) for var_val in params['variableValues']
            )
        else:
            del params['variableValues']
        return params

    def __call__(self, *args, **kwargs) -> typing.Tuple[VariableInstance]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(VariableInstance.load(variable_json) for variable_json in response.json())


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    deserialize_value = QueryParameter('deserializeValue')

    def __init__(self, url: str, id_: str, deserialize_value: bool = False):
        """Get a variable instance.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the variable instance.
        :param deserialize_value: Whether serializable variable values are deserialized on server
                                  side.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.deserialize_value = deserialize_value

    def __call__(self, *args, **kwargs) -> VariableInstance:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return VariableInstance.load(response.json())



# -*- coding: utf-8 -*-

"""This module provides access to the filter REST api of Camunda."""

from __future__ import annotations
import datetime as dt
import dataclasses
import typing

import pycamunda
import pycamunda.task
import pycamunda.variable
import pycamunda.base
from pycamunda.request import QueryParameter, PathParameter, BodyParameter, BodyParameterContainer

URL_SUFFIX = '/filter'


__all__ = ['GetList', 'Count', 'Get', 'Create', 'Update', 'Delete', 'Execute', 'ExecuteCount']


@dataclasses.dataclass
class Query:
    """Data class of query."""

    @classmethod
    def load(cls, **kwargs) -> Query:
        query = cls()
        for key, value in kwargs.items():
            setattr(query, key, value)
        return query


@dataclasses.dataclass
class Properties:
    """Data class of properties."""

    @classmethod
    def load(cls, **kwargs) -> Properties:
        properties = cls()
        for key, value in kwargs.items():
            setattr(properties, key, value)
        return properties


@dataclasses.dataclass
class Filter:
    """Data class of filter as returned by the REST api of Camunda."""
    id_: str
    resource_type: str
    name: str
    owner: str
    query: Query
    properties: Properties
    item_count: int = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Filter:
        return cls(
            id_=data['id'],
            resource_type=data['resourceType'],
            name=data['name'],
            owner=data['owner'],
            query=Query.load(**data['query']),
            properties=Properties.load(**data['properties']),
            item_count=data.get('itemCount', None)
        )


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('filterId')
    resource_type = QueryParameter('resourceType')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    owner = QueryParameter('owner')
    item_count = QueryParameter('itemCount')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'id_': 'filterId',
            'first_name': 'firstName',
            'last_name': 'lastName',
            'email': 'email'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url: str,
        id_: str = None,
        name: str = None,
        name_like: str = None,
        owner: str = None,
        item_count: bool = False,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None
    ):
        """Query for a list of filters using a list of parameters. The size of the result set can be
        retrieved by using the Get Count request.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the filter.
        :param name: Filter by the name of the filter.
        :param name_like: Filter by a substring of the name of the filter.
        :param owner: Filter by the user id of the owner of the filter.
        :param item_count: Return the number of items matched by the respective filters.
        :param sort_by: Sort the results by `id_`, `first_name`, `last_name` or `email` of the user.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.resource_type = 'Task'
        self.name = name
        self.name_like = name_like
        self.owner = owner
        self.item_count = item_count
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[Filter]:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Filter.load(filter_json) for filter_json in response.json())


class Count(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('filterId')
    resource_type = QueryParameter('resourceType')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    owner = QueryParameter('owner')

    def __init__(
        self,
        url: str,
        id_:str = None,
        name: str = None,
        name_like: str = None,
        owner: str = None
    ):
        """Count filters.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the filter.
        :param name: Filter by the name of the filter.
        :param name_like: Filter by a substring of the name of the filter.
        :param owner: Filter by the user id of the owner of the filter.
        """
        super().__init__(url=url + URL_SUFFIX + '/count')
        self.id_ = id_
        self.resource_type = 'Task'
        self.name = name
        self.name_like = name_like
        self.owner = owner

    def __call__(self, *args, **kwargs) -> int:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return response.json()['count']


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('filterId')
    item_count = QueryParameter('itemCount')

    def __init__(self, url: str, id_: str, item_count: bool = False):
        """Query for a filter.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the filter.
        :param item_count: Return the number of items matched by the respective filter.
        """
        super().__init__(url=url + URL_SUFFIX + '/{filterId}')
        self.id_ = id_
        self.item_count = item_count

    def __call__(self, *args, **kwargs) -> Filter:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Filter.load(response.json())


class _Criteria(pycamunda.base.CamundaRequest):

    process_instance_id = BodyParameter('processInstanceId')
    process_instance_business_key = BodyParameter('processInstanceBusinessKey')
    process_instance_business_key_like = BodyParameter('processInstanceBusinessKeyLike')
    process_definition_id = BodyParameter('processDefinitionId')
    process_definition_key = BodyParameter('processDefinitionKey')
    process_definition_key_in = BodyParameter('processDefinitionKeyIn')
    process_definition_name = BodyParameter('processDefinitionName')
    process_definition_name_like = BodyParameter('processDefinitionNameLike')
    case_instance_id = BodyParameter('caseInstanceId')
    case_instance_business_key = BodyParameter('caseInstanceBusinessKey')
    case_instance_business_key_like = BodyParameter('caseInstanceBusinessKeyLike')
    case_definition_id = BodyParameter('caseDefinitionId')
    case_definition_key = BodyParameter('caseDefinitionKey')
    case_definition_name = BodyParameter('caseDefinitionName')
    case_definition_name_like = BodyParameter('caseDefinitionNameLike')
    active = BodyParameter('active')
    activity_instance_id_in = BodyParameter('activityInstanceIdIn')
    execution_id = BodyParameter('executionId')
    assignee = BodyParameter('assignee')
    assignee_in = BodyParameter('assigneeIn')
    assignee_like = BodyParameter('assigneeLike')
    task_owner = BodyParameter('owner')
    candidate_group = BodyParameter('candidateGroup')
    candidate_groups = BodyParameter('candidateGroups')
    candidate_user = BodyParameter('candidateUser')
    involved_user = BodyParameter('involvedUser')
    unassigned = BodyParameter('unassigned')
    delegation_state = BodyParameter('delegationState')
    task_definition_key = BodyParameter('taskDefinitionKey')
    task_definition_key_in = BodyParameter('taskDefinitionKeyIn')
    task_definition_key_like = BodyParameter('taskDefinitionKeyLike')
    task_name = BodyParameter('name')
    task_name_like = BodyParameter('nameLike')
    description = BodyParameter('description')
    description_like = BodyParameter('descriptionLike')
    priority = BodyParameter('priority')
    max_priority = BodyParameter('maxPriority')
    min_priority = BodyParameter('minPriority')
    tenant_id_in = BodyParameter('tenantIdIn')
    without_tenant_id = BodyParameter('withoutTenantId')
    created_before = BodyParameter('createdBefore')
    created_after = BodyParameter('createdAfter')
    due_before = BodyParameter('dueBefore')
    due_after = BodyParameter('dueAfter')
    follow_up_after = BodyParameter('followUpAfter')
    follow_up_before = BodyParameter('followUpBefore')
    follow_up_before_or_not_existent = BodyParameter('followUpBeforeOrNotExistent')
    query = BodyParameterContainer(
        'query',
        process_instance_id,
        process_instance_business_key,
        process_instance_business_key_like,
        process_definition_id,
        process_definition_key,
        process_definition_key_in,
        process_definition_name,
        process_definition_name_like,
        case_instance_id,
        case_instance_business_key,
        case_instance_business_key_like,
        case_definition_id,
        case_definition_key,
        case_definition_name,
        case_definition_name_like,
        active,
        activity_instance_id_in,
        execution_id,
        assignee,
        assignee_in,
        assignee_like,
        task_owner,
        candidate_group,
        candidate_groups,
        candidate_user,
        involved_user,
        unassigned,
        delegation_state,
        task_definition_key,
        task_definition_key_in,
        task_definition_key_like,
        task_name,
        task_name_like,
        description,
        description_like,
        priority,
        max_priority,
        min_priority,
        tenant_id_in,
        without_tenant_id,
        created_before,
        created_after,
        due_before,
        due_after,
        follow_up_after,
        follow_up_before,
        follow_up_before_or_not_existent
    )

    def add_process_instance_criteria(
        self,
        id_: str = ...,
        business_key: str = ...,
        business_key_like: str = ...
    ) -> None:
        """Add criteria that filter by process instance.

        :param id_: Filter by the id of the process instance.
        :param business_key: Filter by the business key of the process instance.
        :param business_key_like: Filter by a substring of the business key of the filter.
        """
        if id_ is not Ellipsis:
            self.process_instance_id = id_
        if business_key is not Ellipsis:
            self.process_instance_business_key = business_key
        if business_key_like is not Ellipsis:
            self.process_instance_business_key_like = business_key_like

    def add_process_definition_criteria(
        self,
        id_: str = ...,
        key: str = ...,
        key_in: typing.Iterable[str] = ...,
        name: str = ...,
        name_like: str = ...
    ) -> None:
        """Add criteria that filter by the process definition.

        :param id_: Filter by the id of the process definition.
        :param key: Filter by the key of the process definition.
        :param key_in: Filter by a substring of the key of the process definition.
        :param name: Filter by the name of the process definition.
        :param name_like: Filter by a substring of the name of the process definition.
        """
        if id_ is not Ellipsis:
            self.process_definition_id = id_
        if key is not Ellipsis:
            self.process_definition_key = key
        if key_in is not Ellipsis:
            self.process_definition_key_in = key_in
        if name is not Ellipsis:
            self.process_definition_name = name
        if name_like is not Ellipsis:
            self.process_definition_name_like = name_like

    def add_case_instance_criteria(
        self,
        id_: str = ...,
        business_key: str = ...,
        business_key_like: str = ...
    ):
        """Add criteria that filter by the case instance.

        :param id_: Filter by the id of the case instance.
        :param business_key: Filter by the business key of the case instance.
        :param business_key_like: Filter by a substring of the business key of the case instance.
        """
        if id_ is not Ellipsis:
            self.case_instance_id = id_
        if business_key is not Ellipsis:
            self.case_instance_business_key = business_key
        if business_key_like is not Ellipsis:
            self.case_instance_business_key_like = business_key_like

        return self

    def add_case_definition_criteria(
        self,
        id_: str = ...,
        key: str = ...,
        name: str = ...,
        name_like: str = ...
    ) -> None:
        """Add criteria that filter by the case definition.

        :param id_: Filter by the id of the case definition.
        :param key: Filter by the key of the case definition.
        :param name: Filter by the name of the case definition.
        :param name_like: Filter by a substring of the name of the case definition.
        """
        if id_ is not Ellipsis:
            self.case_definition_id = id_
        if key is not Ellipsis:
            self.case_definition_key = key
        if name is not Ellipsis:
            self.case_definition_name = name
        if name_like is not Ellipsis:
            self.case_definition_name_like = name_like

    def add_other_criteria(
        self,
        active: bool = ...,
        activity_instance_id_in: typing.Iterable[str] = ...,
        execution_id: str = ...
    ) -> None:
        """Add criteria that filter by active status, activity instance or execution id.

        :param active: Filter only active tasks.
        :param activity_instance_id_in: Filter by activity instance ids.
        :param execution_id: Filter by the execution id.
        """
        if active is not Ellipsis:
            self.active = active or None
        if activity_instance_id_in is not Ellipsis:
            self.activity_instance_id_in = activity_instance_id_in
        if execution_id is not Ellipsis:
            self.execution_id = execution_id

    def add_user_criteria(
        self,
        assignee: str = ...,
        assignee_in: typing.Iterable[str] = ...,
        assignee_like: str = ...,
        task_owner: str = ...,
        candidate_group: str = ...,
        candidate_groups: typing.Iterable[str] = ...,
        candidate_user: str = ...,
        involved_user: str = ...,
        unassigned: bool = ...,
        delegation_state: typing.Union[str, pycamunda.task.DelegationState] = ...
    ) -> None:
        """Add criteria that filter by user.

        :param assignee: Filter by the assignee of the task.
        :param assignee_in: Filter whether assignee of the task is one of multiple ones.
        :param assignee_like: Filter by a substring of the assignee of the task.
        :param task_owner: Filter by the owner of the task.
        :param candidate_group: Filter by the candidate group of the task.
        :param candidate_groups: Filter whether the candidate group of the task is one of multiple
                                 ones.
        :param candidate_user: Filter by the candidate user of the task.
        :param involved_user: TODO
        :param unassigned: Filter only unassigned tasks.
        :param delegation_state: Filter by delegation state.
        """
        if candidate_user is not Ellipsis and (
                candidate_group is not Ellipsis or candidate_groups is not Ellipsis):
            raise pycamunda.PyCamundaException(
                'candidate user and candidate groups must not be both provided.'
            )

        if assignee is not Ellipsis:
            self.assignee = assignee
        if assignee_in is not Ellipsis:
            self.assignee_in = assignee_in
        if assignee_like is not Ellipsis:
            self.assignee_like = assignee_like
        if task_owner is not Ellipsis:
            self.task_owner = task_owner
        if candidate_group is not Ellipsis:
            self.candidate_group = candidate_group
        if candidate_groups is not Ellipsis:
            self.candidate_groups = candidate_groups
        if candidate_user is not Ellipsis:
            self.candidate_user = candidate_user
        if involved_user is not Ellipsis:
            self.involved_user = involved_user
        if unassigned is not Ellipsis:
            self.unassigned = unassigned
        if delegation_state is not Ellipsis:
            self.delegation_state = None
            if delegation_state is not None:
                self.delegation_state = pycamunda.task.DelegationState(delegation_state)

    def add_task_criteria(
        self,
        definition_key: str = ...,
        definition_key_in: typing.Iterable[str] = ...,
        definition_key_like: str = ...,
        task_name: str = ...,
        task_name_like: str = ...,
        description: str = ...,
        description_like: str = ...,
        priority: int = ...,
        max_priority: int = ...,
        min_priority: int = ...,
        tenant_id_in: typing.Iterable[str] = ...,
        without_tenant_id: bool = False
    ) -> None:
        """Add criteria that filter by task.

        :param definition_key: Filter by the definition key of the task.
        :param definition_key_in: Filter whether definition key of the task is one of multiple ones.
        :param definition_key_like: Filter by a substring of the definition key of the task.
        :param task_name: Filter by the name of the task.
        :param task_name_like:  Filter by a substring of the name of the task.
        :param description: Filter by the description of the task.
        :param description_like: Filter by a substring of the description of the task.
        :param priority: Filter by the priority of the task.
        :param max_priority: Filter by a maximum priority of the task.
        :param min_priority: Filter by a minimum priority of the task.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param without_tenant_id: Filter only tasks without tenant id.
        """
        if definition_key is not Ellipsis:
            self.task_definition_key = definition_key
        if definition_key_in is not Ellipsis:
            self.task_definition_key_in = definition_key_in
        if definition_key_like is not Ellipsis:
            self.task_definition_key_like = definition_key_like
        if task_name is not Ellipsis:
            self.task_name = task_name
        if task_name_like is not Ellipsis:
            self.task_name_like = task_name_like
        if description is not Ellipsis:
            self.description = description
        if description_like is not Ellipsis:
            self.description_like = description_like
        if priority is not Ellipsis:
            self.priority = priority
        if max_priority is not Ellipsis:
            self.max_priority = max_priority
        if min_priority is not Ellipsis:
            self.min_priority = min_priority
        if tenant_id_in is not Ellipsis:
            self.tenant_id_in = tenant_id_in
        self.without_tenant_id = without_tenant_id

    def add_datetime_criteria(
        self,
        created_before: dt.datetime = ...,
        created_after: dt.datetime = ...,
        due_before: dt.datetime = ...,
        due_after: dt.datetime = ...,
        follow_up_after: dt.datetime = ...,
        follow_up_before: dt.datetime = ...,
        follow_up_before_or_not_existent: dt.datetime = ...
    ) -> None:
        """Add criteria that filter by datetime. Datetime objects are expected to contain timezone
        information.

        :param created_before: Filter by tasks that were created before the given date.
        :param created_after: Filter by tasks that were created after the given date.
        :param due_before: Filter by tasks where due date has already passed at given date.
        :param due_after: Filter by tasks where due date has not passed at given date.
        :param follow_up_after: Filter by tasks that have a follow up date that has not passed yet
                                at given date.
        :param follow_up_before: Filter by tasks that have a follow up date that has already passed
                                 at given date.
        :param follow_up_before_or_not_existent: Filter by tasks that do not have a follow up date
                                                 or one that has already passed.
        """
        if created_before is not Ellipsis:
            self.created_before = created_before
        if created_after is not Ellipsis:
            self.created_after = created_after
        if due_before is not Ellipsis:
            self.due_before = due_before
        if due_after is not Ellipsis:
            self.due_after = due_after
        if follow_up_after is not Ellipsis:
            self.follow_up_after = follow_up_after
        if follow_up_before is not Ellipsis:
            self.follow_up_before = follow_up_before
        if follow_up_before_or_not_existent is not Ellipsis:
            self.follow_up_before_or_not_existent = follow_up_before_or_not_existent

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)


class Create(_Criteria):

    resource_type = BodyParameter('resourceType')
    name = BodyParameter('name')
    owner = BodyParameter('owner')

    def __init__(self, url: str, name: str, owner: str = None):
        """Create a new filter.

        :param url: Camunda Rest engine URL.
        :param name: Name of the filter.
        :param owner: User id of the owner of the filter.
        """
        super().__init__(url=url + URL_SUFFIX + '/create')
        self.resource_type = 'Task'
        self.name = name
        self.owner = owner

    def __call__(self, *args, **kwargs) -> Filter:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return Filter.load(response.json())


class Update(_Criteria):

    id_ = PathParameter('id')
    resource_type = BodyParameter('resourceType')
    name = BodyParameter('name')
    owner = BodyParameter('owner')

    def __init__(self, url: str, id_: str, name: str = None, owner: str = None):
        """Update a filter.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the filter.
        :param name: Name of the filter.
        :param owner: User id of the owner of the filter.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.resource_type = 'Task'
        self.name = name
        self.owner = owner

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Delete a filter.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the filter.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class Execute(_Criteria):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str, single_result: bool = False):
        """Execute a filter.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the filter.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.single_result = single_result

    @property
    def url(self) -> str:
        return super().url + ('/singleResult' if self.single_result else '/list')

    def body_parameters(self, apply: typing.Callable = ...) -> typing.Dict[str, typing.Any]:
        return super().body_parameters(apply=apply)['query']

    def __call__(
        self, *args, **kwargs
    ) -> typing.Union[pycamunda.task.Task, typing.Tuple[pycamunda.task.Task]]:
        """Send the request."""
        if self.body_parameters():
            response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)
        else:
            response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        if self.single_result:
            return pycamunda.task.Task.load(response.json())
        return tuple(pycamunda.task.Task.load(task_json) for task_json in response.json())


class ExecuteCount(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get the number of results returned by executing a filter.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the filter.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/count')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> int:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return response.json()['count']



# -*- coding: utf-8 -*-

"""This module provides access to the auth REST api of Camunda."""

from __future__ import annotations
import typing
import dataclasses
import enum
import datetime as dt

import pycamunda
import pycamunda.base
import pycamunda.resource
from pycamunda.request import QueryParameter, PathParameter, BodyParameter

URL_SUFFIX = '/auth'


__all__ = ['AuthorizationType', 'GetList', 'Count', 'Get', 'Check', 'Options', 'Update', 'Create',
           'Delete']


class AuthorizationType(enum.IntEnum):
    global_ = 0
    grant = 1
    revoke = 2


@dataclasses.dataclass
class Authorization:
    """Data class of authorization as returned by the REST api of Camunda."""
    id_: str
    type_: AuthorizationType
    permissions: typing.Tuple[str]
    user_id: str
    group_id: str
    resource_type: pycamunda.resource.ResourceType
    resource_id: str
    links: typing.Tuple[pycamunda.resource.Link] = None
    root_process_instance_id: str = None
    removal_time: dt.datetime = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Authorization:
        authorization = cls(
            id_=data['id'],
            type_=data['type'],
            permissions=data['permissions'],
            user_id=data['userId'],
            group_id=data['groupId'],
            resource_type=pycamunda.resource.ResourceType(data['resourceType']),
            resource_id=data['resourceId']
        )
        try:
            authorization.links = tuple(
                pycamunda.resource.Link.load(data=link) for link in data['links']
            )
        except KeyError:
            pass
        try:
            authorization.removal_time = pycamunda.base.from_isoformat(data['removalTime'])
        except KeyError:
            pass
        try:
            authorization.root_process_instance_id = data['rootProcessInstanceId']
        except KeyError:
            pass

        return authorization


@dataclasses.dataclass
class Permission:
    """Data class of permission as returned by the REST api of Camunda."""
    permission_name: str
    resource_name: str
    resource_id: str
    authorized: bool

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Permission:
        return cls(
            permission_name=data['permissionName'],
            resource_name=data['resourceName'],
            resource_id=data['resourceId'],
            authorized=data['authorized']
        )


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    type_ = QueryParameter('type')
    user_id_in = QueryParameter('userIdIn')
    group_id_in = QueryParameter('groupIdIn')
    resource_type = QueryParameter('resourceType')
    resource_id = QueryParameter('resourceId')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'resource_type': 'resourceType',
            'resource_id': 'resourceId'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url: str,
        id_: str = None,
        type_: typing.Union[str, AuthorizationType] = None,
        user_id_in: typing.Iterable[str] = None,
        group_id_in: typing.Iterable[str] = None,
        resource_type: typing.Union[str, pycamunda.resource.ResourceType] = None,
        resource_id: int = None,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None
    ):
        """Query for a list of authorizations using a list of parameters. The size of the result set
        can be retrieved by using the Get Count request.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by the id of the authorization.
        :param type_: Filter by the authorization type.
        :param user_id_in: Filter whether the user id is one of multiple ones.
        :param group_id_in: Filter whether the group id is one of multiple ones.
        :param resource_type: Filter by the resource type.
        :param resource_id: Filter by the resource id.
        :param sort_by: Sort the results by `id_`, `lock_expiration_time, `process_instance_id`,
                        `process_definition_key`, `tenant_id` or `task_priority`.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.type_ = None
        if type_ is not None:
            self.type_ = AuthorizationType(type_)
        self.user_id_in = user_id_in
        self.group_id_in = group_id_in
        self.resource_type = None
        if type_ is not None:
            self.resource_type = pycamunda.resource.ResourceType(resource_type)
        self.resource_id = resource_id
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[Authorization]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Authorization.load(auth_json) for auth_json in response.json())


class Count(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    type_ = QueryParameter('type')
    user_id_in = QueryParameter('userIdIn')
    group_id_in = QueryParameter('groupIdIn')
    resource_type = QueryParameter('resourceType')
    resource_id = QueryParameter('resourceId')

    def __init__(
        self,
        url: str,
        id_: str = None,
        type_: typing.Union[str, AuthorizationType] = None,
        user_id_in: typing.Iterable[str] = None,
        group_id_in: typing.Iterable[str] = None,
        resource_type: typing.Union[str, pycamunda.resource.ResourceType] = None,
        resource_id: int = None,
    ):
        """Get the size of the result returned by the Get List request.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by the id of the authorization.
        :param type_: Filter by the authorization type.
        :param user_id_in: Filter whether the user id is one of multiple ones.
        :param group_id_in: Filter whether the group id is one of multiple ones.
        :param resource_type: Filter by the resource type.
        :param resource_id: Filter by the resource id.
        """
        super().__init__(url=url + URL_SUFFIX + '/count')
        self.id_ = id_
        self.type_ = None
        if type_ is not None:
            self.type_ = AuthorizationType(type_)
        self.user_id_in = user_id_in
        self.group_id_in = group_id_in
        self.resource_type = None
        if type_ is not None:
            self.resource_type = pycamunda.resource.ResourceType(resource_type)
        self.resource_id = resource_id

    def __call__(self, *args, **kwargs) -> int:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return int(response.json()['count'])


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get an auth.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the authorization.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> Authorization:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Authorization.load(data=response.json())


class Check(pycamunda.base.CamundaRequest):

    permission_name = QueryParameter('permissionName')
    permission_value = QueryParameter('permissionValue')
    resource_name = QueryParameter('resourceName')
    resource_type = QueryParameter('resourceType')
    resource_id = QueryParameter('resourceId')

    def __init__(
        self,
        url: str,
        permission_name: str,
        permission_value: int,
        resource_name: str,
        resource_type: typing.Union[str, pycamunda.resource.ResourceType],
        resource_id: str = None
    ):
        """Check the authorization of the currently authenticated user.

        :param url: Camunda Rest engine URL.
        :param permission_name: Name of the permission to check.
        :param permission_value: Value of the permission to check for.
        :param resource_name: Name of the resource to check for.
        :param resource_type: Type of the resource to check for.
        :param resource_id: Id of the resource to check for.
        """
        super().__init__(url=url + URL_SUFFIX + '/check')
        self.permission_name = permission_name
        self.permission_value = permission_value
        self.resource_name = resource_name
        self.resource_type = resource_type
        self.resource_id = resource_id

    def __call__(self, *args, **kwargs) -> Permission:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Permission.load(data=response.json())


class Options(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str = None):
        """Get a list of options the currently authenticated user can perform on the authorization
        resource.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the authorization
        """
        super().__init__(url=url + URL_SUFFIX + '{path}')
        self.id_ = id_

    @property
    def url(self):
        return self._url.format(path='/{id}'.format(id=self.id_) if self.id_ is not None else '')

    def __call__(self, *args, **kwargs) -> pycamunda.resource.ResourceOptions:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.OPTIONS, *args, **kwargs)

        return pycamunda.resource.ResourceOptions.load(data=response.json())


class Create(pycamunda.base.CamundaRequest):

    type_ = BodyParameter('type')
    permissions = BodyParameter('permissions')
    user_id = BodyParameter('userId')
    group_id = BodyParameter('groupId')
    resource_type = BodyParameter('resourceType')
    resource_id = BodyParameter('resourceId')

    def __init__(
        self,
        url: str,
        type_: typing.Union[int, AuthorizationType],
        permissions: typing.Iterable[str],
        resource_type: typing.Union[str, pycamunda.resource.ResourceType],
        resource_id: str,
        user_id: str = None,
        group_id: str = None
    ):
        """Create an auth.

        :param url: Camunda Rest engine URL.
        :param type_: Id of the authorization.
        :param permissions: Permissions provided by this authorization. A permission be 'READ' or
                            'CREATE' for example.
        :param user_id: Id of the user this authorization is for. The value '*' means all users.
        :param group_id: Id of the group this authorization is for.
        :param resource_type: Resource type this authorization is for.
        :param resource_id: Id of the resource. The value '*' means all instances of a resource.
        """
        super().__init__(url=url + URL_SUFFIX + '/create')
        self.type_ = type_
        self.permissions = permissions
        self.user_id = user_id
        self.group_id = group_id
        self.resource_type = resource_type
        self.resource_id = resource_id

    def __call__(self, *args, **kwargs) -> Authorization:
        """Send the request."""
        assert (self.user_id is not None) != (self.group_id is not None), (
            'Either \'user_id\' or \'group_id\' has to be provided, not both.'
        )
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return Authorization.load(data=response.json())


class Update(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    permissions = BodyParameter('permissions')
    user_id = BodyParameter('userId')
    group_id = BodyParameter('groupId')
    resource_type = BodyParameter('resourceType')
    resource_id = BodyParameter('resourceId')

    def __init__(
        self,
        url: str,
        id_: str,
        permissions: typing.Iterable[str],
        resource_type: typing.Union[str, pycamunda.resource.ResourceType],
        resource_id: str,
        user_id: str = None,
        group_id: str = None
    ):
        """Update an auth.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the authorization.
        :param permissions: Permissions provided by this authorization. A permission be 'READ' or
                            'CREATE' for example.
        :param user_id: Id of the user this authorization is for. The value '*' means all users.
        :param group_id: Id of the group this authorization is for.
        :param resource_type: Resource type this authorization is for.
        :param resource_id: Id of the resource. The value '*' means all instances of a resource.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.permissions = permissions
        self.user_id = user_id
        self.group_id = group_id
        self.resource_type = resource_type
        self.resource_id = resource_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        assert (self.user_id is not None) != (self.group_id is not None), (
            'Either \'user_id\' or \'group_id\' has to be provided, not both.'
        )
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Delete an auth.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the authorization.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)



# -*- coding: utf-8 -*-

"""This module provides access to the resource REST api of Camunda."""

from __future__ import annotations
import dataclasses
import typing
import enum


__all__ = ['ResourceType']


@dataclasses.dataclass
class Link:
    """Data class of link as returned by the REST api of Camunda."""
    method: str
    href: str
    rel: str

    @classmethod
    def load(cls, data) -> Link:
        return Link(
            method=data['method'],
            href=data['href'],
            rel=data['rel'],
        )


@dataclasses.dataclass
class ResourceOptions:
    """Data class of resource options as returned by the REST api of Camunda."""
    links: typing.Tuple[Link]

    def __iter__(self):
        return (link for link in self.links)

    def __len__(self):
        return len(self.links)

    @classmethod
    def load(cls, data) -> ResourceOptions:
        return ResourceOptions(
            links=tuple(Link(**link) for link in data['links'])
        )


class ResourceType(enum.IntEnum):
    application = 0
    user = 1
    group = 2
    group_membership = 3
    authorization = 4
    filter_ = 5
    process_definition = 6
    task = 7
    process_instance = 8
    deployment = 9
    decision_definition = 10
    tenant = 11
    tenant_membership = 12
    batch = 13
    decision_requirements_definition = 14
    report = 15
    dashboard = 16
    user_operation_log_category = 17
    optimize = 18
    historic_task = 19
    historic_process_instance = 20



# -*- coding: utf-8 -*-

"""This module provides access to the incident REST api of Camunda."""

from __future__ import annotations
import datetime as dt
import dataclasses
import enum
import typing

import pycamunda.variable
import pycamunda.base
from pycamunda.request import QueryParameter, PathParameter

URL_SUFFIX = '/incident'


__all__ = ['IncidentType', 'Get', 'GetList', 'Resolve']


class IncidentType(enum.Enum):
    failed_job = 'failedJob'
    failed_external_task = 'failedExternalTask'


@dataclasses.dataclass
class IncidentTypeCount:
    """Data class of incident type count as returned by the REST api of Camunda."""
    incident_type: IncidentType
    incident_count: int

    @classmethod
    def load(cls, data) -> IncidentTypeCount:
        return cls(
            incident_type=IncidentType(data['incidentType']),
            incident_count=data['incidentCount']
        )


@dataclasses.dataclass
class Incident:
    """Data class of incident as returned by the REST api of Camunda."""
    id_: str
    process_definition_id: str
    process_instance_id: str
    execution_id: str
    incident_type: IncidentType
    activity_id: str
    cause_incident_id: str
    root_cause_incident_id: str
    configuration: str
    tenant_id: str
    incident_message: str
    job_definition_id: str
    incident_timestamp: dt.datetime = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Incident:
        incident = cls(
            id_=data['id'],
            process_definition_id=data['processDefinitionId'],
            process_instance_id=data['processInstanceId'],
            execution_id=data['executionId'],
            incident_type=IncidentType(data['incidentType']),
            activity_id=data['activityId'],
            cause_incident_id=data['causeIncidentId'],
            root_cause_incident_id=data['rootCauseIncidentId'],
            configuration=data['configuration'],
            tenant_id=data['tenantId'],
            incident_message=data['incidentMessage'],
            job_definition_id=data['jobDefinitionId']
        )
        if data['incidentTimestamp'] is not None:
            incident.incident_timestamp = pycamunda.base.from_isoformat(
                data['incidentTimestamp']
            )

        return incident


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get an incident.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the incident.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> Incident:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Incident.load(response.json())


class GetList(pycamunda.base.CamundaRequest):

    incident_id = QueryParameter('incidentId')
    incident_type = QueryParameter('incidentType')
    incident_message = QueryParameter('incidentMessage')
    process_definition_id = QueryParameter('processDefinitionId')
    process_definition_key_in = QueryParameter('processDefinitionKeyIn')
    process_instance_id = QueryParameter('processInstanceId')
    execution_id = QueryParameter('executionId')
    activity_id = QueryParameter('activityId')
    cause_incident_id = QueryParameter('causeIncidentId')
    root_cause_incident_id = QueryParameter('rootCauseIncidentId')
    configuration = QueryParameter('configuration')
    tenant_id_in = QueryParameter('tenantIdIn')
    job_definition_id_in = QueryParameter('jobDefinitionIdIn')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'incident_id': 'incidentId',
            'incident_message': 'incidentMessage',
            'incident_timestamp': 'incidentTimestamp',
            'incident_type': 'incidentType',
            'execution_id': 'executionId',
            'activity_id': 'activityId',
            'process_instance_id': 'processInstanceId',
            'process_definition_id': 'processDefinitionId',
            'cause_incident_id': 'causeIncidentId',
            'root_cause_incident_id': 'rootCauseIncidentId',
            'configuration': 'configuration',
            'tenant_id': 'tenantId'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )

    def __init__(
        self,
        url: str,
        incident_id: str = None,
        incident_type: str = None,
        incident_message: str = None,
        process_definition_id: str = None,
        process_definition_key_in: typing.Iterable[str] = None,
        process_instance_id: str = None,
        execution_id: str = None,
        activity_id: str = None,
        cause_incident_id: str = None,
        root_cause_incident_id: str = None,
        configuration: str = None,
        tenant_id_in: typing.Iterable[str] = None,
        job_definition_id_in: typing.Iterable[str] = None,
        sort_by: str = None,
        ascending: bool = True
    ):
        """Get a list of incidents.

        :param url: Camunda Rest engine URL.
        :param incident_id: Filter by incident id.
        :param incident_type: Filter by incident type. Valid values are 'failedJob' and
                              'failedExternalTask'
        :param incident_message: Filter by incident message.
        :param process_definition_id: Filter by process definition id.
        :param process_definition_key_in: Filter whether the process definition key is one of
                                          multiple ones.
        :param process_instance_id: Filter by process instance id.
        :param execution_id: Filter by execution id.
        :param activity_id: Filter by activity id.
        :param cause_incident_id: Filter by cause incident id.
        :param root_cause_incident_id: Filter by root cause incident id.
        :param configuration: Filter by configuration.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param job_definition_id_in: Filter whether the job definition id is one of multiple ones.
        :param sort_by: Sort the results by 'incident_id', 'incident_message', incident_timestamp',
                        'incident_type', 'execution_id', 'activity_id', 'process_instance_id',
                        'process_definition_id', 'cause_incident_id', 'root_cause_incident_id',
                        'configuration' or 'tenant_id'.
        :param ascending: Sort order.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.incident_id = incident_id
        self.incident_type = incident_type
        self.incident_message = incident_message
        self.process_definition_id = process_definition_id
        self.process_definition_key_in = process_definition_key_in
        self.process_instance_id = process_instance_id
        self.execution_id = execution_id
        self.activity_id = activity_id
        self.cause_incident_id = cause_incident_id
        self.root_cause_incident_id = root_cause_incident_id
        self.configuration = configuration
        self.tenant_id_in = tenant_id_in
        self.job_definition_id_in = job_definition_id_in
        self.sort_by = sort_by
        self.ascending = ascending

    def __call__(self, *args, **kwargs) -> typing.Tuple[Incident]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Incident.load(incident_json) for incident_json in response.json())


class Resolve(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Resolve an incident.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the incident.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)



# -*- coding: utf-8 -*-

"""This module provides access to the message REST api of Camunda."""

from __future__ import annotations
import dataclasses
import enum
import typing

import pycamunda.processinst
import pycamunda.execution
import pycamunda.variable
import pycamunda.base
from pycamunda.request import BodyParameter

URL_SUFFIX = '/message'


__all__ = ['ResultType', 'CorrelateSingle', 'CorrelateAll']


class ResultType(enum.Enum):
    process_definition = 'ProcessDefinition'
    execution = 'Execution'


@dataclasses.dataclass
class MessageCorrelationResult:
    """Data class of message correlation result as returned by the REST api of Camunda."""
    result_type: ResultType
    process_instance: pycamunda.processinst.ProcessInstance = None
    execution: pycamunda.execution.Execution = None
    variables: typing.Tuple[pycamunda.variable.Variable] = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> MessageCorrelationResult:

        message_result = cls(result_type=ResultType(data['resultType']))

        if message_result.result_type == ResultType.process_definition:
            message_result.process_instance = pycamunda.processinst.ProcessInstance.load(
                data['processInstance']
            )
        elif message_result.result_type == ResultType.execution:
            message_result.execution = pycamunda.execution.Execution.load(data['execution'])

        try:
            variables = data['variables']
        except KeyError:
            pass
        else:
            message_result.variables = tuple(
                pycamunda.variable.Variable.load(variable_json) for variable_json in variables
            )

        return message_result


class _Correlate(pycamunda.base.CamundaRequest):

    message_name = BodyParameter('messageName')
    business_key = BodyParameter('businessKey')
    tenant_id = BodyParameter('tenantId')
    without_tenant_id = BodyParameter('withoutTenantId', provide=pycamunda.base.value_is_true)
    process_instance_id = BodyParameter('processInstanceId')
    correlation_keys = BodyParameter('correlationKeys')
    local_correlation_keys = BodyParameter('localCorrelationKeys')
    process_variables = BodyParameter('processVariables')
    process_variables_local = BodyParameter('processVariablesLocal')
    all_ = BodyParameter('all')
    result_enabled = BodyParameter('resultEnabled')
    variables_in_result_enabled = BodyParameter('variablesInResultEnabled')

    def __init__(
        self,
        url: str,
        message_name: str,
        all_: bool,
        business_key: str = None,
        tenant_id: str = None,
        without_tenant_id: bool = False,
        process_instance_id: str = None,
        result_enabled: bool = False,
        variables_in_result_enabled: bool = False
    ):
        """Correlate a message to one or multiple entities. Entities are executions and process
        definitions.

        :param url: Camunda Rest engine URL.
        :param message_name: Name of the message to correlate.
        :param all_: Whether to correlate the message to all or just one entity.
        :param business_key: Correlate only to executions that belong to a process instance with the
                             provided business key.
        :param tenant_id: Correlate only to entities which belong to the provided tenant.
        :param without_tenant_id: Whether to correlate only to entities that belong to no tenant.
        :param process_instance_id: Correlate only to a specific process instance.
        :param result_enabled: Whether to return message correlation results.
        :param variables_in_result_enabled: Whether the returned message correlation results contain
                                            process instance variables.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.message_name = message_name
        self.all_ = all_
        self.business_key = business_key
        self.tenant_id = tenant_id
        self.without_tenant_id = without_tenant_id
        self.process_instance_id = process_instance_id
        self.result_enabled = result_enabled
        self.variables_in_result_enabled = variables_in_result_enabled

        self.correlation_keys = {}
        self.local_correlation_keys = {}
        self.process_variables = {}
        self.process_variables_local = {}

    def add_correlation_key(self, name: str, value: typing.Any, type_: str = None) -> None:
        """Add a correlation key. Used for correlation of process instances that wait for incoming
        messages. Only global process instance variables are considered.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Type of the variable.
        """
        self.correlation_keys[name] = {'value': value, 'type': type_}

    def add_local_correlation_key(self, name: str, value: typing.Any, type_: str = None) -> None:
        """Add a correlation key. Used for correlation of process instances that wait for incoming
        messages. Only variables in the execution scope are considered.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Type of the variable.
        """
        self.local_correlation_keys[name] = {'value': value, 'type': type_}

    def add_process_variable(
        self,
        name: str,
        value: typing.Any,
        type_: str = None,
        value_info: typing.Mapping = None
    ) -> None:
        """Add variables to the process after correlating the message.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.process_variables[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

    def add_local_process_variable(
        self,
        name: str,
        value: typing.Any,
        type_: str = None,
        value_info: typing.Mapping = None
    ) -> None:
        """Add local variables to the process after correlating the message.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.process_variables_local[name] = {
            'value': value, 'type': type_, 'valueInfo': value_info
        }

    def __call__(self, *args, **kwargs) -> typing.Tuple[MessageCorrelationResult]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return tuple(
            MessageCorrelationResult.load(data=result_json) for result_json in response.json()
        )


class CorrelateSingle(_Correlate):

    def __init__(
        self,
        url: str,
        message_name: str,
        business_key: str = None,
        tenant_id: str = None,
        without_tenant_id: bool = False,
        process_instance_id: str = None,
        result_enabled: bool = False,
        variables_in_result_enabled: bool = False
    ):
        """Correlate a message to one or multiple entities. Entities are executions and process
        definitions.

        :param url: Camunda Rest engine URL.
        :param message_name: Name of the message to correlate.
        :param all_: Whether to correlate the message to all or just one entity.
        :param business_key: Correlate only to executions that belong to a process instance with the
                             provided business key.
        :param tenant_id: Correlate only to entities which belong to the provided tenant.
        :param without_tenant_id: Whether to correlate only to entities that belong to no tenant.
        :param process_instance_id: Correlate only to a specific process instance.
        :param result_enabled: Whether to return message correlation results.
        :param variables_in_result_enabled: Whether the returned message correlation results contain
                                            process instance variables.
        """
        super().__init__(
            url=url,
            message_name=message_name,
            all_=False,
            business_key=business_key,
            tenant_id=tenant_id,
            without_tenant_id=without_tenant_id,
            process_instance_id=process_instance_id,
            result_enabled=result_enabled,
            variables_in_result_enabled=variables_in_result_enabled
        )


class CorrelateAll(_Correlate):

    def __init__(
        self,
        url: str,
        message_name: str,
        business_key: str = None,
        tenant_id: str = None,
        without_tenant_id: bool = False,
        process_instance_id: str = None,
        result_enabled: bool = False,
        variables_in_result_enabled: bool = False
    ):
        """Correlate a message to one or multiple entities. Entities are executions and process
        definitions.

        :param url: Camunda Rest engine URL.
        :param message_name: Name of the message to correlate.
        :param all_: Whether to correlate the message to all or just one entity.
        :param business_key: Correlate only to executions that belong to a process instance with the
                             provided business key.
        :param tenant_id: Correlate only to entities which belong to the provided tenant.
        :param without_tenant_id: Whether to correlate only to entities that belong to no tenant.
        :param process_instance_id: Correlate only to a specific process instance.
        :param result_enabled: Whether to return message correlation results.
        :param variables_in_result_enabled: Whether the returned message correlation results contain
                                            process instance variables.
        """
        super().__init__(
            url=url,
            message_name=message_name,
            all_=True,
            business_key=business_key,
            tenant_id=tenant_id,
            without_tenant_id=without_tenant_id,
            process_instance_id=process_instance_id,
            result_enabled=result_enabled,
            variables_in_result_enabled=variables_in_result_enabled
        )
